
def check_house_availability(arg_name, arg_value, collection):
    """
    This method will check if the house in the refit dataset exist and will return a flag either True or False 

    Parameters 
    ----------
    arg_name :  string
                name of parameter
    arg_value : (any)
                value of the target parameter
    target_datatype : (any)
                      data type of the target parameter

    returns: boolean
                True if all the validations for target parameter are validated correct otherwise False
    """
    try:
        if arg_value in collection:
            return True
        
        elif arg_value == None:
            print(f"NoneTypeError: Argument value provided is 'None'")
            return False
        
        elif isinstance(arg_value, str):
            print(f"TypeError: String not accepted. Expected value of datatype <class 'int'>")
            return False

        else:
            print(f"{arg_name} = {arg_value} does not exist in the provided dataset.")
            return False

    except Exception as e:
        print("Error occured in check_availability method due to ", e)
        

def check_correct_datatype(arg_name, arg_value, target_datatype):
    """
    This method will check all the validations and will return a flag either True or False depending on the correct validations

    Parameters 
    ----------
    arg_name :  string
                name of parameter
    arg_value : (any)
                value of the target parameter
    target_datatype : (any)
                      data type of the target parameter

    returns: boolean
                True if all the validations for target parameter are validated correct otherwise False
    """
    try:
        if isinstance(arg_value, target_datatype):
            return True
        
        elif arg_value == None:
            print(f"NoneTypeError: Argument '{arg_name}' cannot be 'None' and it accepts datatype {target_datatype})")
            return False

        else:
            print(f"TypeError: Argument '{arg_name}' accepts datatype {target_datatype}")
            return False

    except Exception as e:
        print("Error occured in check_correct_datatype method due to ", e)

        
def check_list_validations(arg_name, arg_value, member_datatype):
    """
    This method will check all the validations and will return a flag either True or False depending on the correct validations

    Parameters 
    ----------
    arg_name :  string
                name of parameter
    arg_value : (any)
                value of the target parameter
    member_datatype : (any)
                      data type of the members in the list

    returns: boolean
                True if all the validations for target list are validated correct otherwise False
    """
    try:
        if check_correct_datatype(arg_name, arg_value, list):
            if len(arg_value)!=0:
                return True
            else:
                print(f"Error: Empty list. Please specify some values using the argument '{arg_name}' <class 'list'>: ({member_datatype})")
                return False

    except Exception as e:
        print("Error occured in check_correct_datatype method due to ", e)




import numpy as np
import pandas as pd
from refit_loader.utilities import convert_object2timestamps


def __generate_activity_report(df, target_appliance, threshold):
    """
    This method will return the durations or events when the appliance was active (on) 
    
    Parameters 
    ----------
    data : pandas.core.frame.DataFrame
            example dataframe is of the following format            
            {
                'time': pandas.core.indexes.datetimes.DatetimeIndex
                    timestamps as index identifying every data row
                'aggregate': numpy.int64
                    aggregated power consumption of all appliances in the specified house (watts)
                'kettle': numpy.int64
                    kettle power consumption in the specified house (watts)
            }
    target_appliance : string
            name of the target appliance (may be the name of the column targeted)
    threshold : float
            value of threshold for raw samples of power consumption, above this threshold appliance is consider active
        
    returns: pandas.core.frame.DataFrame
            dataframe is of the following format            
            {
                'Activity_Start': pandas._libs.tslibs.timestamps.Timestamp
                    start of the duration/event 
                'Activity_End': pandas._libs.tslibs.timestamps.Timestamp 
                    end of the duration/event
                'Duration': float
                    minutes of active appliance (using the method = convert_timestamps2minutes to convert timestamps to minutes)
            }
    """
    try: 
        duration_start = []
        duration_end = []
        duration_size = []

        if isinstance(df.index, object):
            df.index = convert_object2timestamps(df.index)

        df_tmp = df[[target_appliance]].copy()
        mask = df[target_appliance] > threshold
        df_tmp['mask'] = (mask)
        df_tmp['cum_sum'] = (~mask).cumsum()
        df_tmp = df_tmp[df_tmp['mask'] == True]
        df_tmp = df_tmp.groupby(['cum_sum', str(df.index.name)]).first()

        for x in df_tmp.index.unique(level='cum_sum'):
            d = df_tmp.loc[(x)].reset_index()
            duration_start.append(d.iloc[0][str(df.index.name)])
            duration_end.append(d.iloc[-1][str(df.index.name)])
            duration_size.append(duration_end[-1] - duration_start[-1])
        durations = (pd.Series(duration_size)) / np.timedelta64(1, 'm')
        return pd.DataFrame({'Activity_Start': duration_start, 'Activity_End': duration_end, 'Duration': durations})
    
    except Exception as e:
        print("Exception raised in generate_activity_report() method = ", e)


        
def get_activities(data, target_appliance=None, threshold=None):
    """
    This method will call the generate_activity_report for every dataframe to compute the durations of the active appliance and append to either a dictionary or returns back the single dataframe
    
    Parameters 
    ----------
    data : dict or pandas.core.frame.DataFrame
                 dictionary =
                         contains dataframes of multiple houses where key represents the house number (int) and value represents (pandas.core.frame.DataFrame)
                         every dataframe includes the aggregate consumption and power consumption of a specific appliance in float values
                         with timestamps as index
                 pandas.core.frame.DataFrame  =
                         single dataframe of a single house that includes the aggregate consumption and power consumption of a specific appliance in float values
                         with timestamps as index
                         example dataframe is of the following format            
                            {
                                'time': pandas.core.indexes.datetimes.DatetimeIndex
                                    timestamps as index identifying every data row
                                'aggregate': numpy.int64
                                    aggregated power consumption of all appliances in the specified house (watts)
                                'kettle': numpy.int64
                                    kettle power consumption in the specified house (watts)
                            }
                     
    target_appliance : string
            name of the target appliance (may be the name of the column targeted)
    threshold : float
            value of threshold for raw samples of power consumption, above this threshold appliance is consider active 
        
    returns: dict or pandas.core.frame.DataFrame
                dictionary =
                        contains dataframes of multiple houses where key represents the house number (int) and value represents (pandas.core.frame.DataFrame)
                pandas.core.frame.DataFrame  =
                        dataframe is of the following format            
                            {
                                'Activity_Start': pandas._libs.tslibs.timestamps.Timestamp
                                    start of the duration/event 
                                'Activity_End': pandas._libs.tslibs.timestamps.Timestamp 
                                    end of the duration/event
                                'Duration': float
                                    minutes of active appliance (using the method = convert_timestamps2minutes to convert timestamps to minutes)
                            }
    """
    try:
        if threshold is None:
                threshold = 0.0
        
        if isinstance(data, dict):
            house_activities = {}
            for key, df in data.items():
                
                if target_appliance is None:
                    if len(df.columns)==2:
                        target_appliance = df.columns[-1]
                    else:
                        raise Exception(f"Please specify target appliance {df.columns}")
                    
                print(f"Estimating active durations of House {key}: {target_appliance}")
                house_activities.update({key: __generate_activity_report(df, target_appliance, threshold)})

            return house_activities
        
        elif isinstance(data, pd.core.frame.DataFrame):
            
            if target_appliance is None:
                if len(data.columns)==2:
                    target_appliance = data.columns[-1]
                else:
                    raise Exception(f"Please specify target_appliance \n {data.columns}")
            print(f"Estimating active durations of: {target_appliance}")
            return __generate_activity_report(data, target_appliance, threshold)

        else:
            print(f"Provided data should be of type <dict> or <pandas.core.frame.DataFrame> and not {type(data)}.")
    
    except Exception as e:
        print("Exception raised in get_activities() method = ", e)


from os.path import isdir, dirname, abspath
from inspect import currentframe, getfile, getsourcefile
from os import getcwd
from sys import getfilesystemencoding
import pandas as pd
import numpy as np


def get_module_directory():
    # Taken from http://stackoverflow.com/a/6098238/732596
    path_to_this_file = dirname(getfile(currentframe()))
    if not isdir(path_to_this_file):
        encoding = getfilesystemencoding()
        path_to_this_file = dirname(unicode(__file__, encoding))
    if not isdir(path_to_this_file):
        abspath(getsourcefile(lambda _: None))
    if not isdir(path_to_this_file):
        path_to_this_file = getcwd()
    assert isdir(path_to_this_file), path_to_this_file + ' is not a directory'
    return path_to_this_file


def convert_timestamps2minutes(tstamps: pd.Series):
    try:
        return tstamps / np.timedelta64(1, 'm')

    except Exception as e:
        print("Exception raised in generating convert_timestamps2minutes() method = ", e)


def convert_object2timestamps(objects: pd.Series):
    try:
        return pd.to_datetime(objects)

    except Exception as e:
        print("Exception raised in generating convert_objects2timestamps() method = ", e)



import pandas as pd
import os, pickle
from sklearn.preprocessing import StandardScaler, MinMaxScaler, RobustScaler
import math
import dask.dataframe as dd
from refit_loader.modules.parser import refit_parser
from refit_loader.modules.validations import check_house_availability, check_list_validations, check_correct_datatype
from refit_loader.modules.active_durations import get_activities
from refit_loader.utilities import convert_object2timestamps, get_module_directory
    
class __Loader:
    """
    Interface that loads all the data into the memory
    """

    def __init__(self):
        try:
            pass
            
        except Exception as e:
            print("Error occured in initialization of _Loader interface due to ", e)
                
        finally:
            pass
        
    @staticmethod
    def _load_file():
        raise NotImplementedError    


        
class CSV_Loader(__Loader):
    """
    Class that loads all the data into the memory using different methods
    """
    def __init__(self):
        try:
            super().__init__()
        
        except Exception as e:
            print("Error occured in initialization of CSV_Loader class due to ", e)
                
        finally:
            pass
        
    @staticmethod
    def _load_file(csv_file_path,
                   index_column_name=None,
                   _nrows=None,
                   _iterator=True,
                   _chunksize=100000):
        try:
            tp = pd.read_csv(csv_file_path, nrows=_nrows, index_col=index_column_name, iterator=_iterator, chunksize=_chunksize) ## loading data in chunks reduces 90 percent execution time 
            df = pd.concat(tp, ignore_index=False)
            df.info(verbose=False, memory_usage="deep")
            return df  
        
        except Exception as e:
            print("Error occured in _load_file method of CSV_Loader class due to ", e)
    
    @staticmethod
    def _load_files_via_dask(_data_folder,
                             _files_format,
                             _buildings):
        try:
            ls = {}
            print(f"\nLoading specified buildings: {_buildings}")
            for i in _buildings:
                ls.update({i: dd.read_csv(f"{_data_folder}{i}{_files_format}")})
            return ls
        
        except Exception as e:
            print("Error occured in _load_file_via_dask method of CSV_Loader class due to ", e)
    

    
class REFIT_Loader(CSV_Loader):
    """
    Class that loads all the data into the memory using DASK
    """
    def __init__(self, data_path):
        try:
            super().__init__()
        
        except Exception as e:
            print("Error occured in initialization of REFIT_Loader class due to ", e)
                
        finally:
            # self.__config = get_config_from_json(description="refit_loader configuration", config_file='../refit_loader/config.json')
            self.__config = {'DATA_FOLDER': os.path.join(os.getcwd(), data_path), 'DATA_TYPE':'.csv', 'README_FILE': os.path.join(get_module_directory(), 'metadata', 'REFIT_README.txt'),
                             'REFIT_HOUSES': [1,2,3,4,5,6,7,8,9,10,11,12,13,15,16,17,18,19,20,21]}
            self.__collective_dataset = CSV_Loader._load_files_via_dask(_data_folder=self.__config['DATA_FOLDER']+'House_',
                                                                _files_format=self.__config['DATA_TYPE'],
                                                                _buildings=self.__config['REFIT_HOUSES'])
            self.__keys_of_appliances = refit_parser(self.__config['README_FILE'])
            for house_number in self.__collective_dataset:
                cols = [header.lower() for header in self.__keys_of_appliances[str(house_number)]]
                self.__collective_dataset[house_number] = self.__collective_dataset[house_number].rename(columns={"Time": "time", "Unix": "unix", "Aggregate": cols[0], "Appliance1":cols[1], "Appliance2":cols[2],
                                                                                                      "Appliance3":cols[3], "Appliance4":cols[4], "Appliance5":cols[5],"Appliance6":cols[6], "Appliance7":cols[7],
                                                                                                      "Appliance8":cols[8], "Appliance9":cols[9]})
                self.__collective_dataset[house_number].index = self.__collective_dataset[house_number]['time']
                self.__collective_dataset[house_number] = self.__collective_dataset[house_number].drop('time', axis=1)
                
    def get_appliance_names(self, house: int):
        """
        This method will return the names of the available appliances in the specified house number

        Parameters 
        ----------
        house : int 
                number of a specific house e.g, house=2 for 'house_2.csv' 

        returns: list
                 contains names of the available appliances in the specified house
        """
        try:
            if check_correct_datatype(arg_name='house', arg_value=house, target_datatype=int):
                if check_house_availability(arg_name='House Number', arg_value=house, collection=self.__collective_dataset.keys()):
                    print(f"Fetching appliances for house = {house}")
                    return [name for name in self.__collective_dataset[house].columns]
        except Exception as e:
            print("Error occured in get_appliance_names method of REFIT_Loader due to ", e)
                
    def get_house_data(self, house: int):
        """
        This method will return the dataframe of the specified house number from refit dataset

        Parameters 
        ----------
        house : int 
                number of a specific house e.g, house=2 for 'house_2.csv' 

        returns: pandas.core.frame.DataFrame
                dataframe is of the following format            
                {
                    'time': pandas.core.indexes.datetimes.DatetimeIndex
                        timestamps as index identifying every data row
                    'unix': numpy.int64
                        timestamps in unix
                    'aggregate': numpy.int64
                        aggregated power consumption of all appliances in the sepcified house
                    
                    *** appliances and their consumption values in numpy.int64 varies house wise ***
                }
        """
        try:                
            if check_correct_datatype(arg_name='house', arg_value=house, target_datatype=int):
                if check_house_availability(arg_name='House Number', arg_value=house, collection=self.__collective_dataset.keys()):
                    print(f"Loading data for house = {house}")
                    data = self.__collective_dataset[house].compute() 
                    data.index = convert_object2timestamps(data.index)
                    data = data.loc[:, data.columns != "unix"].astype(float)
                    return data
        
        except Exception as e:
            print("Error occured in get_house_data method of REFIT_Loader due to ", e)
    
    def get_appliance_data(self, appliance: str, houses=None):
        """
        This method will return RefitData object that can let user access data in dictionary format as well can access some transformation methods

        Parameters 
        ----------
        appliance : string
                name of the target appliance (name of the column targeted in the specified house/s)
        house : list
                contains numbers of a specific houses e.g, house=2 for 'house_2.csv' 

        returns: RefitData object
                .data = to access data in a dictionary format
                        dictionary contains dataframes of multiple houses where key represents the house number (int) and value represents (pandas.core.frame.DataFrame)
                        dataframe is of the following format            
                        {
                            'time': pandas.core.indexes.datetimes.DatetimeIndex
                                timestamps as index identifying every data row
                            'unix': numpy.int64
                                timestamps in unix
                            'aggregate': numpy.int64
                                aggregated power consumption of all appliances in the sepcified house

                            *** appliances and their consumption values in numpy.int64 varies house wise ***
                        }
        """
        try:
            self.__data = {}
            if check_correct_datatype(arg_name='appliance', arg_value=appliance, target_datatype=str):
                target_appliance = appliance.lower()
            if houses == None:
                houses=list(self.__collective_dataset.keys())
                
            print(f"Loading data for appliance {target_appliance.upper()} ...")
            if check_list_validations(arg_name='houses', arg_value=houses, member_datatype='int'):
                for house_number in houses:
                    if check_house_availability(arg_name='House Number', arg_value=house_number, collection=self.__collective_dataset.keys()):
                        if target_appliance in self.__collective_dataset[house_number].columns:
                            if house_number not in self.__data.keys():
                                print(f"Fetching {target_appliance.upper()} data for House {house_number}")
                                data = self.__collective_dataset[house_number][['aggregate', target_appliance]].compute()
                                data.index = convert_object2timestamps(data.index)
                                data = data.astype(float)
                                self.__data.update({house_number: data})
                        else:
                            print(f"Appliance '{target_appliance.upper()}' does not exist in house {house_number}.")

            return RefitData(self.__data)
                
        except Exception as e:
            print("Error occured in get_appliance_data method of REFIT_Loader due to ", e)


class RefitData():
    """
    Class that loads the provided data after computing DASK dataframes and provides different methods for transformations
    """
    def __init__(self, data):
        try:
            self.data = data
            self.splits = dict()
        
        except Exception as e:
            print("Error occured in initialization of RefitData class due to ", e)
                
        finally:
            pass
    
    def resample(self, sampling_period='8s', window_limit=3.0, fill_value=0.0):
        """
        This method will return RefitData object that can let user access data in dictionary format as well can access some transformation methods

        Parameters 
        ----------
        house : int 
                number of a specific house e.g, house=2 for 'house_2.csv' 
        sampling_period: string
                         set the sampling rate in a string format e.g, '8s' means 8 seconds
        window_limit : float
                        set the window size in minutes to forward fill last value
        fill_value: float
                    set the value with which remaining np.nans are filled 

        returns: RefitData object (updated)
                .data = to access data in a dictionary format
                        dictionary contains dataframes of multiple houses where key represents the house number (int) and value represents (pandas.core.frame.DataFrame)
                        dataframe is of the following format            
                        {
                            'time': pandas.core.indexes.datetimes.DatetimeIndex
                                timestamps as index identifying every data row
                            'unix': numpy.int64
                                timestamps in unix
                            'aggregate': numpy.int64
                                aggregated power consumption of all appliances in the sepcified house

                            *** appliances and their consumption values in numpy.int64 varies house wise ***
                        }
        """
        try:
            self.__sampling_period = sampling_period
            self.__fill_value = fill_value
            self.__window_limit= int(window_limit*60)

            for house_number in self.data.keys():
                print(f"Resampling for house number: ", house_number)
#                     target_appliance = self.data[house_number].columns[-1]
                appliance_data = self.data[house_number]
#                     appliance_data = appliance_data.resample('1s').mean().dropna()
                appliance_data = appliance_data.resample('1s').asfreq()
                appliance_data.fillna(method='ffill', axis=0, inplace=True, limit=self.__window_limit)
                appliance_data.fillna(axis=0, inplace=True, value=self.__fill_value)
                appliance_data = appliance_data.resample(self.__sampling_period).median()
                # appliance_data.dropna(inplace = True)
                self.data.update({house_number: appliance_data})
            print("Updating data with resampled dataset...")

        except Exception as e:
            print("Error occured in resample method of REFIT_Data due to ", e)
            
            
    def subset_data(self, no_of_days=5, threshold = None ):

            """
            This method will create different and smaller versions of the training, validation and testing subsets from the collective_data

            Parameters 
            ----------
            no_of_days: int
                            number of days with active appliance time/durations

            returns: RefitData object (updated)
                .active_data = to access subset_data in a dictionary format
                        dictionary contains dataframes of multiple houses where key represents the house number (int) and value represents (pandas.core.frame.DataFrame)
                        dataframe is of the following format            
                        {
                            'time': pandas.core.indexes.datetimes.DatetimeIndex
                                timestamps as index identifying every data row
                            'aggregate': numpy.int64
                                aggregated power consumption of all appliances in the sepcified house
                            *** target_appliance : numpy.int64
                                power consumption of target appliances in the sepcified house
                        }
            """
            self.__no_of_days = no_of_days
            for house_number, value in self.data.items():
                print(f"Subetting dataset with {self.__no_of_days} days of most activities for House {house_number}")
                activities = get_activities(value, threshold=threshold)
                date_wise_activities = activities.groupby([activities['Activity_Start'].dt.date]).sum()
                time_indices = date_wise_activities.sort_values('Duration', ascending=False).head(self.__no_of_days).index
                df_outer = pd.DataFrame()
                for version, time_indx in enumerate(time_indices):
                    df_outer = pd.concat([df_outer, value.loc[str(time_indx)]])
                df_outer.sort_index(inplace=True)
                self.data.update({house_number: df_outer})
            print("Updating data with selected active appliance activities...")
            
            
    def get_proportioned_data(self, target_houses, splits_proportion):
        """
        """
        try:
            self.__target_houses = target_houses
            self.__proportion = splits_proportion

            train_split, val_split, test_split = pd.DataFrame(), pd.DataFrame(), pd.DataFrame()

            if (target_houses['TRAIN'].sort() == target_houses['VALIDATE'].sort()) and (target_houses['TRAIN'].sort() == target_houses['TEST'].sort()) :

                for house_number in self.__target_houses['TRAIN']:
                    df = self.data[house_number]
                    __train_end = df.index[math.floor(self.__proportion['TRAIN_PERCENT'] * len(df))]
                    __val_end = df.index[math.floor((self.__proportion['TRAIN_PERCENT'] + self.__proportion['VALIDATE_PERCENT']) * len(df))]
                    train_split = pd.concat([train_split, df[:__train_end]])
                    val_split = pd.concat([val_split, df[__train_end:__val_end]])
                    test_split = pd.concat([test_split, df[__val_end:]])

            elif (target_houses['TRAIN'].sort() != target_houses['VALIDATE'].sort()) and (target_houses['TRAIN'].sort() != target_houses['TEST'].sort()):

                for house_number in self.__target_houses['TRAIN']:
                    df = self.data[house_number]
                    __train_end = df.index[math.floor(self.__proportion['TRAIN_PERCENT'] * len(df))]
                    train_split = pd.concat([train_split, df[:__train_end]])
                for house_number in self.__target_houses['VALIDATE']:
                    df = self.data[house_number]
                    __val_end = df.index[math.floor(self.__proportion['VALIDATE_PERCENT'] * len(df))]
                    val_split = pd.concat([val_split, df[:__val_end]])
                for house_number in self.__target_houses['TEST']:
                    df = self.data[house_number]
                    __test_end = df.index[math.floor(self.__proportion['TEST_PERCENT'] * len(df))]
                    test_split = pd.concat([test_split, df[:__test_end]])

            elif target_houses['TRAIN'].sort() == target_houses['VALIDATE'].sort():

                for house_number in self.__target_houses['TRAIN']:
                    df = self.data[house_number]
                    __train_end = df.index[math.floor(self.__proportion['TRAIN_PERCENT'] * len(df))]
                    __val_end = df.index[math.floor((self.__proportion['TRAIN_PERCENT'] + self.__proportion['VALIDATE_PERCENT']) * len(df))]
                    train_split = pd.concat([train_split, df[:__train_end]])
                    val_split = pd.concat([val_split, df[__train_end:__val_end]])
                for house_number in self.__target_houses['TEST']:
                    df = self.data[house_number]
                    __test_end = df.index[math.floor(self.__proportion['TEST_PERCENT'] * len(df))]
                    test_split = pd.concat([test_split, df[:__test_end]])

            else:
                pass

            train_split.sort_index(inplace=True)
            val_split.sort_index(inplace=True)
            test_split.sort_index(inplace=True)
            self.splits.update({'TRAIN': train_split, 'VALIDATE': val_split, 'TEST': test_split})
            print("Updating splits with specified proportion from every target house...")

        except Exception as e:
            print("Error occured in get_proportioned_data method due to ", e)

    def normalize(self, scaler, scalars_directory, training=False ):
        """
        This method will standardize the values of the provided dataset. It will compute the scalars on trainset and save them to provided path which can later be used to standarize validation set

        Parameters
        ----------
        train_flag : bool
                to indicate the provided dataframe is of training dataset
        df: pandas.core.frame.DataFrame
                dataframe is of the following format
                {
                    'time': pandas.core.indexes.datetimes.DatetimeIndex
                        timestamps as index identifying every data row
                    'aggregate': numpy.int64
                        aggregated power consumption of all appliances in the sepcified house
                    * any target appliance (vary) *: numpy.int64
                        aggregated power consumption of all appliances in the sepcified house
                }

        returns: scaled_df: pandas.core.frame.DataFrame

        """
        try:
            if scaler == 'Standard' or scaler == 'standard':
                X_scaler, y_scaler = StandardScaler(), StandardScaler()

            elif scaler == 'MinMax' or scaler == 'minmax':
                X_scaler, y_scaler = MinMaxScaler(), MinMaxScaler()

            elif scaler == 'Robust' or scaler == 'robust':
                X_scaler, y_scaler = RobustScaler(), RobustScaler()

            else:
                print('Argument \'scaler\' value is undefined! Setting it to standard scaling....')
                X_scaler, y_scaler = StandardScaler(), StandardScaler()

            if not os.path.exists(scalars_directory):
                os.makedirs(f'{scalars_directory}/', )  # if not then create folder

            if not training:
                for house_number in self.data.keys():
                    print(f"Normalizing for house number: ", house_number)
                    __appliance_data = self.data[house_number]
                    X_array = __appliance_data['aggregate'].values
                    y_array = __appliance_data[__appliance_data.columns[-1]].values

                    X_scaler.fit(X_array.reshape(-1, 1))
                    y_scaler.fit(y_array.reshape(-1, 1))
                    pickle.dump(X_scaler, open(os.path.join(scalars_directory, f'X_scaler_house_{house_number}.sav'), 'wb'))
                    pickle.dump(y_scaler, open(os.path.join(scalars_directory, f'y_scaler_house_{house_number}.sav'), 'wb'))
                    X = X_scaler.transform(X_array.reshape(-1, 1)).flatten()
                    y = y_scaler.transform(y_array.reshape(-1, 1)).flatten()
                    __normalized_df = pd.DataFrame({'time': __appliance_data.index, 'aggregate': X, f"{__appliance_data.columns[-1]}":y}).set_index('time')
                    self.data.update({house_number: __normalized_df})
                print("Updating data with normalized dataset...")

            else:
                print('Normalization is being performed for training a model. Scalers will be computed/fit considering the train_split and using those scalers, all splits will be normalized/transformed.')

                if len(self.splits) == 0:
                    raise Exception("Call the function get_proportioned_data to specify the training, validation and testing splits, in order to perform normalization for training.")
                else:
                    X_array = self.splits['TRAIN']['aggregate'].values
                    y_array = self.splits['TRAIN'][self.splits['TRAIN'].columns[-1]].values

                    X_scaler.fit(X_array.reshape(-1, 1))
                    y_scaler.fit(y_array.reshape(-1, 1))

                    pickle.dump(X_scaler, open(os.path.join(scalars_directory, f'X_scaler.sav'), 'wb'))
                    pickle.dump(y_scaler, open(os.path.join(scalars_directory, f'y_scaler.sav'), 'wb'))

                    X_train = X_scaler.transform(X_array.reshape(-1, 1)).flatten()
                    y_train = y_scaler.transform(y_array.reshape(-1, 1)).flatten()

                    __train_df = pd.DataFrame({'time': self.splits['TRAIN'].index, 'aggregate': X_train,
                                               f"{self.splits['TRAIN'].columns[-1]}": y_train}).set_index('time')

                    X_array = self.splits['VALIDATE']['aggregate'].values
                    y_array = self.splits['VALIDATE'][self.splits['VALIDATE'].columns[-1]].values

                    X_val = X_scaler.transform(X_array.reshape(-1, 1)).flatten()
                    y_val = y_scaler.transform(y_array.reshape(-1, 1)).flatten()

                    X_array = self.splits['TEST']['aggregate'].values
                    y_array = self.splits['TEST'][self.splits['TEST'].columns[-1]].values

                    X_test = X_scaler.transform(X_array.reshape(-1, 1)).flatten()
                    y_test = y_scaler.transform(y_array.reshape(-1, 1)).flatten()

                    __val_df = pd.DataFrame({'time': self.splits['VALIDATE'].index, 'aggregate': X_val, f"{self.splits['VALIDATE'].columns[-1]}": y_val}).set_index('time')
                    __test_df = pd.DataFrame({'time': self.splits['TEST'].index, 'aggregate': X_test,f"{self.splits['TEST'].columns[-1]}": y_test}).set_index('time')

                    print("Updating splits with normalized data splits...")
                    self.splits = {'TRAIN': __train_df, 'VALIDATE': __val_df, 'TEST': __test_df}

        except Exception as e:
            print("Error occured in normalize method of REFIT_Data due to ", e)






def refit_parser(readme_file):
    """
    This method will return the dictionary of the appliances, parsed from README file of refit data

    Parameters 
    ----------
    readme_file : txt file 
                    original README file of refit datase

    returns: pandas.Dictionary
                    dictionary contains the all the appliances in a specific house as mentioned in README file
    """
    try:
        print(f'Parsing the readme file specified: {readme_file}')
        with open(readme_file) as f:
            content = f.readlines()
            content.append('\n')
    
        ls = {}
        for i, s in enumerate(content):

            if '0.Aggregate,' in s:
                
                keys, appliances = [], []
                house = content[i-1].split()[1]

                for indx in range(0, 3):
                    target = [value.split('.') for value in [value for value in content[i+indx].split(',') if value != '\n']]
                    for t in target:
                        if len(t) > 2: ##### one comma missing in house 5 caused issue
                            appliances.append(t[1])
                            appliances.append(t[2])
                        else:
                            appliances.append(t[1])
                ls.update({house: [item.split('\n')[0].replace(" ", "") for item in appliances]})
        return ls
    
    except Exception as e:
        print("Error occured in parser method due to ", e)